from __future__ import print_function
import sys
import copy
import time
import os
import serial

from ... import PY3
from ...uarm import UArm
from ..list_uarms import get_uarm_port_cli
from ...protocol import *
from ..firmware  import flash,get_uarm_port_cli, download, default_config
from ...util import printf


if PY3:
    izip = zip
else:
    from itertools import izip


__version__ = "2.0.0"

if getattr(sys, 'frozen', False):
    FROZEN_APP = False
else:
    FROZEN_APP = True

if FROZEN_APP:
    application_path = os.path.dirname(__file__)
else:
    application_path = os.path.dirname(sys.executable)

default_calibration_hex = 'calibration.hex'



def calibrate(port_name):
      # Flash the calibration firmware
    sp = serial.Serial(port=port_name, baudrate=115200)
    READY = '[STEP]READY'
    START = '[STEP]START'
    COMPLETE = '[STEP]COMPLETE'
    if PY3:
        ready_msg = bytes(READY, encoding='ascii')
        start_msg = bytes(START, encoding='ascii')
        complete_msg = bytes(COMPLETE, encoding='ascii')
    else:
        ready_msg = READY
        start_msg = START
        complete_msg = COMPLETE
    while True:
        if get_serial_line(sp).startswith(ready_msg):  # Waiting For READY MESSAGE
            break
        time.sleep(0.01)
    while True:
        if get_serial_line(sp).startswith(start_msg):  # Waiting For START MESSAGE
            break
        time.sleep(0.01)
    while True:
        if get_serial_line(sp).startswith(complete_msg):  # Waiting For COMPLETED MESSAGE
            break
        time.sleep(0.01)


def get_serial_line(sp):
    line = sp.readline()
    if line != b'':
        printf(line)
    return line



def read_manual_offset(uarm):
    address = OFFSET_START_ADDRESS
    read_manual_offset = []
    for i in range(4):
        read_manual_offset.append(round(uarm.get_rom_data(address, EEPROM_DATA_TYPE_FLOAT), 2))
        address += 4
    return read_manual_offset


def read_linear_offset(uarm):
    linear_offset_template = {"INTERCEPT": 0.00, "SLOPE": 0.00}
    intercept_address = LINEAR_INTERCEPT_START_ADDRESS
    slope_address = LINEAR_SLOPE_START_ADDRESS
    linear_offset_data = []
    for i in range(4):
        linear_offset= copy.deepcopy(linear_offset_template)
        linear_offset['INTERCEPT'] = round(
            uarm.get_rom_data(intercept_address, EEPROM_DATA_TYPE_FLOAT), 2)
        linear_offset['SLOPE'] = round(uarm.get_rom_data(slope_address, EEPROM_DATA_TYPE_FLOAT), 2)
        linear_offset_data.append(linear_offset)
        intercept_address += 4
        slope_address += 4
    return linear_offset_data


def read_completed_flag(uarm, flag_type):
    if flag_type == CALIBRATION_FLAG:
        if uarm.get_rom_data(CALIBRATION_FLAG) == CONFIRM_FLAG:
            return True
        else:
            return False
    elif flag_type == CALIBRATION_LINEAR_FLAG:
        if uarm.get_rom_data(CALIBRATION_LINEAR_FLAG) == CONFIRM_FLAG:
            return True
        else:
            return False
    elif flag_type == CALIBRATION_SERVO_FLAG:
        if uarm.get_rom_data(CALIBRATION_SERVO_FLAG) == CONFIRM_FLAG:
            return True
        else:
            return False

def exit_fun():
    try:
        input("\nPress Enter to Exit...")
        sys.exit(0)
    except Exception:
        pass



def main(args):
    if args.port:
        port_name = args.port
    else:
        port_name = get_uarm_port_cli()

    if args.debug:
        debug = True
    else:
        debug = False

    if args.check:
        uarm = UArm(port_name=port_name, debug=debug)
        printf("All Calibration: {}".format("COMPLETED" if read_completed_flag(uarm, CALIBRATION_FLAG) else "NOT COMPLETED"))
        printf("Linear Calibration: {}".format("COMPLETED" if read_completed_flag(uarm, CALIBRATION_LINEAR_FLAG) else "NOT COMPLETED"))
        printf("Manual Calibration: {}".format("COMPLETED" if read_completed_flag(uarm, CALIBRATION_SERVO_FLAG) else "NOT COMPLETED"))
        for linear_offset,manual_offset, i in izip(read_linear_offset(uarm), read_manual_offset(uarm), range(4)):
            print ("Servo {} INTERCEPT: {}, SLOPE: {}, MANUAL: {}".format(i,linear_offset['INTERCEPT'], linear_offset['SLOPE'], manual_offset))
    else:
        uarm = UArm(port_name=port_name, debug=debug)
        calibration_completed_flag = read_completed_flag(uarm, CALIBRATION_FLAG)
        printf("All Calibration: {}".format("COMPLETED" if calibration_completed_flag else "NOT COMPLETED"))
        if calibration_completed_flag:
            if PY3:
                choice = input("Calibration Completed, Are you sure to continue? Yes, please press Y\n")
            else:
                choice = raw_input("Calibration Completed, Are you sure to continue? Yes, please press Y\n")
            if choice.lower() == 'y':
                if PY3:
                    input("Please don't quit until calibration completed. Continue please press Enter.\n")
                else:
                    raw_input("Please don't quit until calibration completed. Continue please press Enter.\n")
                firmware_path = os.path.join(os.getcwd(), default_config['filename'])
                download(default_config['download_url'], firmware_path)
                flash(port_name, default_calibration_hex)
                calibrate(port_name)
                flash(port_name, firmware_path)
            else:
                print ("Exit")
        else:
            if PY3:
                input("Please don't quit until calibration completed. Continue please press Enter.\n")
            else:
                raw_input("Please don't quit until calibration completed. Continue please press Enter.\n")
            firmware_path = os.path.join(os.getcwd(), default_config['filename'])
            download(default_config['download_url'], firmware_path)
            flash(port_name, default_calibration_hex)
            calibrate(port_name)
            flash(port_name, firmware_path)


if __name__ == '__main__':
    try:
        import argparse
        parser = argparse.ArgumentParser()
        parser.add_argument("-p", "--port", help="specify port number")
        parser.add_argument("-d", "--debug", help="Open Debug Message", action="store_true")
        parser.add_argument("-c", "--check", help="Check the calibrate offset values", action="store_true")
        args = parser.parse_args()
        main(args)
    except Exception as e:
        print (str(e))



from __future__ import print_function
from __future__ import division

__version__ = '1.0.5'
__author__  = 'Alex Tan'
'''
This Tool is for uArm firmware flashing. Also support download firmware online
'''

import sys
from serial.tools import list_ports
import os
import io
import platform
import subprocess
import json
from os.path import expanduser

if sys.version > '3':
    PY3 = True
else:
    PY3 = False

if PY3:
    import urllib.request
else:
    import urllib2

default_config = {
    "filename": "firmware.hex",
    "hardware_id": "USB VID:PID=0403:6001",
    "download_url": "http://download.ufactory.cc/firmware/firmware_dev.hex",
    "download_flag": False
}

home_dir = os.path.join(expanduser("~"), ".uarm", "")
ua_dir = os.path.join(home_dir, "assistant")

if getattr(sys, 'frozen', False):
    FROZEN_APP = True
elif __file__:
    FROZEN_APP = False




def resourcePath(relative_path):
    """ Get absolute path to resource, works for dev and for PyInstaller """
    if hasattr(sys, '_MEIPASS'):
        return os.path.join(sys._MEIPASS, relative_path)
    if FROZEN_APP:
        return os.path.join(os.path.abspath("."),  "pyuarm" , "tools", "firmware",  relative_path)
    else:
        return os.path.join(os.path.dirname(__file__), relative_path)


def exit_fun():
    try:
        input("\nPress Enter to Exit...")
    except Exception:
        pass


def uarm_ports(hardware_id=default_config["hardware_id"]):
    uarm_ports = []
    for i in list_ports.comports():
        if i.hwid[0:len(hardware_id)] == hardware_id:
            uarm_ports.append(i[0])
    return uarm_ports


def get_uarm_port_cli():
    ports = uarm_ports()
    if len(ports) > 1:
        i = 1
        for port in ports:
            print(("[{}] - {}".format(i, port)))
            i += 1
        port_index = input("Please Choose the uArm Port: ")
        uarm_port = ports[int(port_index) - 1]
        return uarm_port
    elif len(ports) == 1:
        return ports[0]
    elif len(ports) == 0:
        print("No uArm ports is found.")
        return None


def download(url, filepath):
    try:
        if PY3:
            u = urllib.request.urlopen(url)
            fileTotalbytes = u.length
        else:
            u = urllib2.urlopen(url)
            fileTotalbytes = int(u.info().getheaders("Content-Length")[0])
        print ("writing to {}, file size: {} bytes ".format(filepath, str(fileTotalbytes)))

        data_blocks = []
        total=0

        while True:
            block = u.read(1024)
            data_blocks.append(block)
            total += len(block)
            hash = ((60*total)//fileTotalbytes)
            per =total / fileTotalbytes
            if PY3:
                print("[{}{}] {:.0%}".format('#' * hash, ' ' * (60-hash), per), end="\r")
            else:
                print("[{}{}] {:.0%}".format('#' * hash, ' ' * (60 - hash), per), end="\r")
            if not len(block):
                print ("\nCompleted!")
                break

        data=b''.join(data_blocks) #had to add b because I was joining bytes not strings
        u.close()


        with open(filepath, "wb") as f:
                f.write(data)
    except Exception as e:
        print ("Error: " + str(e))


def gen_flash_cmd(port, firmware_path, avrdude_path=None):
    global avrdude_bin, avrdude_conf, error_description, cmd
    avrdude_conf = ""
    if avrdude_path is None:
        avrdude_path = resourcePath('avrdude')
    if port:
        if platform.system() == 'Darwin':
            avrdude_bin = os.path.join(avrdude_path, 'mac', 'avrdude')
            avrdude_conf = os.path.join(avrdude_path, 'mac', 'avrdude.conf')
            avrdude_conf = '-C' + avrdude_conf
            error_description = "built-in avrdude is not working, Please install avrdude..."

        elif platform.system() == 'Windows':
            avrdude_bin = os.path.join(avrdude_path, 'windows',  'avrdude.exe')
            avrdude_conf = os.path.join(avrdude_path,'windows',  'avrdude.conf')
            avrdude_conf = '-C' + avrdude_conf
            error_description = "built-in avrdude is not working, Please install winavr..."

        elif platform.system() == 'Linux':
            avrdude_bin = "avrdude"
            error_description = "Please install avrdude first..."
        cmd = [avrdude_bin, avrdude_conf, '-v', '-patmega328p', '-carduino', '-P' + port, '-b115200', '-D',
               '-Uflash:w:{0}:i'.format(firmware_path)]
    return cmd


def flash(port, firmware_path, avrdude_path=None):
        cmd = gen_flash_cmd(port,firmware_path,avrdude_path)
        print((' '.join(cmd)))
        try:
            subprocess.call(cmd)
        except OSError as e:
            print(("Error occurred: error code {0}, error msg: {1}".format(str(e.errno), e.strerror)))


class FlashFirmware:

    def __init__(self):
        if FROZEN_APP:
            self.application_path = os.path.dirname(__file__)
        else:
            self.application_path = os.path.dirname(sys.executable)

        if PY3:
            config_path = os.path.join(self.application_path, 'config.json')
        else:
            config_path = os.path.join(self.application_path, 'config.json').decode("utf8").encode("latin-1")
        if not os.path.exists(config_path):
            print ("config.json not found, use default setting: " + str(default_config))
            print ("\n")
            config = default_config
        else:
            print("Loading config file:" + str(config_path))
            with io.open(config_path, "r", encoding="utf-8") as data_file:
                config = json.load(data_file)

        if PY3:
            self.firmware_path = os.path.join(self.application_path, config['filename'])
        else:
            self.firmware_path = os.path.join(self.application_path, config['filename'].decode("utf8").encode("latin-1"))
        self.hardware_id = config['hardware_id']
        self.firmware_url = config['download_url']
        self.download_flag = config['download_flag']

    def flash_firmware(self):
        flash(port=get_uarm_port_cli(), firmware_path=self.firmware_path)

    def download_firmware(self):
        print ("Downloading firmware.hex...")
        download(self.firmware_url, self.firmware_path)


def main(args):
    if args.port:
        port_name = args.port
    else:
        port_name = get_uarm_port_cli()
    if args.path:
        firmware_path = args.path
    else:
        if FROZEN_APP:
            firmware_path = os.path.join(os.path.dirname(sys.executable), default_config['filename'])
        else:
            firmware_path = os.path.join(os.path.dirname(__file__), default_config['filename'])

    if args.download:
        download(default_config['download_url'], firmware_path)

    avrdude_path = None
    if not FROZEN_APP:
        avrdude_path = os.path.join(os.path.dirname(__file__), 'avrdude')

    flash(port_name, firmware_path,avrdude_path)

if __name__ == '__main__':
    try:
        import argparse

        parser = argparse.ArgumentParser()
        parser.add_argument("--port", help="specify port number")
        parser.add_argument("--path", help="firmware path")
        parser.add_argument("-d", "--download",
                        help="download firmware from {}".format(default_config['download_url']),
                        action="store_true")
        args = parser.parse_args()
        main(args)
    except SystemExit:
        exit_fun()





import logging
from .version import __version__

ERROR = 2
INFO  = 1
DEBUG = 0

global pylogger

def get_default_logger(debug=False):
    if debug:
        logging_level = logging.DEBUG
    else:
        logging_level = logging.INFO
    logger = logging.getLogger('pyuarm')
    logger.setLevel(logging_level)
    formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    ch.setLevel(logging_level)
    logger.addHandler(ch)
    return logger

def init_logger(logger):
    """
    initialize global logger
    :param debug: if True, Turn on the logger debug
    :return:
    """
    global pylogger
    pylogger = logger
    printf('pyuarm version: ' + __version__)

def printf(msg, type=INFO):
    """
    global print log function
    :param msg:
    :param type:
    :return:
    """
    if type == INFO:
        pylogger.info(msg)
    elif type == DEBUG:
        pylogger.debug(msg)
    elif type == ERROR:
        pylogger.error(msg)


class UArmConnectException(Exception):
    def __init__(self, errno, message=None):
        """
        uArm Connect Exception
        :param errno: 0 Unable to connect uArm, 1 unknown firmware version, 2 unsupported uArm Frimware version
        :param message:
        """
        if message is None:
            self.message = ""
        else:
            self.message = message
        self.errno = errno
        if self.errno == 0:
            self.error = "Unable to connect uArm"
        elif self.errno == 1:
            self.error = "Unknown Firmware Version"
        elif self.errno == 2:
            self.error = "Unsupported uArm Firmware Version"
        else:
            self.error = "Not Defined Error"

    def __str__(self):
        return repr(self.error + "-" + self.message)


# parameters

## SERVO NUMBER INDEX
SERVO_BOTTOM = 0
SERVO_LEFT = 1
SERVO_RIGHT = 2
SERVO_HAND = 3

## CALIBRATION FLAG
CONFIRM_FLAG = 0x80
CALIBRATION_FLAG = 10
CALIBRATION_LINEAR_FLAG = 11
CALIBRATION_SERVO_FLAG = 12
CALIBRATION_STRETCH_FLAG = 13

## OFFSET EEPROM ADDRESS
LINEAR_INTERCEPT_START_ADDRESS = 70
LINEAR_SLOPE_START_ADDRESS = 50
OFFSET_START_ADDRESS = 30
OFFSET_STRETCH_START_ADDRESS = 20

SERIAL_NUMBER_ADDRESS = 100

## EEPROM DATA TYPE INDEX
EEPROM_DATA_TYPE_BYTE = 1
EEPROM_DATA_TYPE_INTEGER = 2
EEPROM_DATA_TYPE_FLOAT = 4

## PROTOCOL MESSAGE
READY                   = "@1"
OK                      = "OK"
SET_POSITION            = "G0 X{} Y{} Z{} F{}"
SET_POSITION_RELATIVE   = "G204 X{} Y{} Z{} F{}"
SIMULATION              = "M222 X{} Y{} Z{} P0"
GET_FIRMWARE_VERSION    = "P203"
GET_HARDWARE_VERSION    = "P202"
SET_ANGLE               = "G202 N{} V{}"
# SET_RAW_ANGLE           = "sSerN{}V{}"
STOP_MOVING             = "G203"
SET_PUMP                = "M231 V{}"
GET_PUMP                = "P231"
SET_GRIPPER             = "M232 V{}"
GET_GRIPPER             = "P232"
ATTACH_SERVO            = "M201 N{}"
DETACH_SERVO            = "M202 N{}"
GET_COOR                = "P220"
GET_ANGLE               = "P200"
# GET_RAW_ANGLE           = "gSer"
GET_IS_MOVE             = "M200"
GET_TIP_SENSOR          = "P233"
SET_BUZZER              = "M210 F{} T{}"
SET_POLAR               = "G201 S{} R{} H{} F{}"
GET_POLAR               = "P221"
GET_EEPROM              = "M211 N0 A{} T{}"
SET_EEPROM              = "M212 N0 A{} T{} V{}"
GET_ANALOG              = "P241 N{}"
GET_DIGITAL             = "P240 N{}"


import re
__version__ = '2.2.5.1'
support_versions = ['2.2']


def is_a_version(version):
    version_pattern = re.compile(r'\d+\.\d+\.\d+\w*')
    if version_pattern.match(version):
        return True
    else:
        return False


def is_supported_version(version):
    pattern = re.compile(r'\d+\.\d+')
    major_version = pattern.match(version).group()
    for v in support_versions:
        if major_version == v:
            return True
    return False



#!/usr/bin/env python
#
# This is a module that help user to control your uArm.
# Please use -h to list all parameters of this script

# This file is part of pyuarm. https://github.com/uArm-Developer/pyuarm
# (C) 2016 UFACTORY <developer@ufactory.cc>


from cmd import Cmd
from .list_uarms import get_uarm_port_cli, uarm_ports
from ..uarm import UArm
from .. import util

version = "0.1.4"


class UArmCmd(Cmd):

    help_msg = "Shortcut:" + "\n"
    help_msg += "Quit: " + "Ctrl + D"
    help_msg += ", or input: " + "quit" + "\n"
    help_msg += "Clear Screen: " + "Ctrl + L"

    ON_OFF = ['on', 'off']

    FIRMWARE =['version', 'force', 'upgrade']

    SERVO_STATUS = ['attach', 'detach']

    prompt = ">>> "
    intro = "Welcome to use uArm Command Line - v{}\n"\
        .format(version)

    intro += help_msg
    intro += "\n\n"
    intro += "Input help for more usage"

    # doc_header = "documented commands:"
    ruler = '-'

    uarm = None

    def __init__(self, port=None,debug=False, *args, **kwargs):
        Cmd.__init__(self, *args, **kwargs)
        self.connect(port=port,debug=debug)

    def __is_connected(self):
        if self.uarm is None:
            print ("No uArm is connected, please use connect")
            return False
        else:
            if self.uarm.is_connected():
                return True
            else:
                print ("No uArm is connected, please use connect command")
                return False

    def do_connect(self, arg):
        """
        connect, Open uArm Port, if more than one uArm Port found, will prompt options to choose.
        Please connect to uArm before you do any control action
        """
        if len(arg) != 0:
            self.connect(arg)
        elif len(arg) == 0:
            self.connect()

    def connect(self, port=None, debug=False):
        """
        connect uArm.
        :param port:
        :return:
        """

        if self.uarm is None:
            if port is not None:
                self.uarm = UArm(port_name=port, debug=debug)
            else:
                ports = uarm_ports()
                if len(ports) > 1:
                    uarm_port = get_uarm_port_cli()
                    try:
                        self.uarm = UArm(uarm_port)
                    except util.UArmConnectException as e:
                        print ("uArm Connect failed, {}".format(str(e)))
                elif len(ports) == 1:
                    self.uarm = UArm()
                elif len(ports) == 0:
                    print ("No uArm ports is found.")
        else:
            if self.uarm.is_connected():
                print ("uArm is already connected, port: {}".format(self.uarm.port.device))
            else:
                if self.uarm.connect():
                    print ("uArm port: {} is reconnected")

    def do_disconnect(self,arg):
        """
        disconnect, Release uarm port.
        """
        if self.uarm is not None:
            if self.uarm.is_connected():
                self.uarm.disconnect()

    def do_set_position(self, arg):
        """
        set_position, move to destination coordinate.
        format: set_position X Y Z or move_to X Y Z S
        X,Y,Z unit millimeter, S means Speed, unit mm/s
        eg. set_position 100 200 150
        """
        if self.__is_connected():
            values = arg.split(' ')
            if len(values) == 3:
                result = self.uarm.set_position(int(values[0]), int(values[1]), int(values[2]))
                msg = "succeed" if result else "failed"
                print (msg)
            elif len(values) == 4:
                result = self.uarm.set_position(int(values[0]), int(values[1]), int(values[2]), int(values[3]))
                msg = "succeed" if result else "failed"
                print (msg)

    def do_sp(self, args):
        """
        same with set_position
        """
        self.do_set_position(args)

    def do_get_position(self, arg):
        """
        get_position get current coordinate
        """
        if self.__is_connected():
            coords = self.uarm.get_position()
            print ("Current coordinate: X:{} Y:{} Z:{}".format(*coords))

    def do_pump(self, arg):
        """
        pump
        turn on/off pump
        """
        if self.__is_connected():
            if arg == 'on':
                result = self.uarm.set_pump(True)
                print ("succeed" if result else "failed")
            elif arg == 'off':
                result = self.uarm.set_pump(False)
                print ("succeed" if result else "failed")
            elif arg == '':
                print ("please input argument: {}".format(','.join(self.ON_OFF)))
            else:
                print ("Command not found {}".format(arg))

    def complete_pump(self, text, line, begidx, endidx):
        if not text:
            completions = self.ON_OFF[:]
        else:
            completions = [f
                           for f in self.ON_OFF
                           if f.startswith(text)
                           ]
        return completions

    def do_sim(self, arg):
        """
        sim
        format: sim X Y Z
        validate the coordinate.
        eg. sim 100 200 100
        succeed
        """
        if self.__is_connected():
            values = arg.split(' ')
            if len(values) == 3:
                result = self.uarm.get_simulation(int(values[0]), int(values[1]), int(values[2]))
                msg = "succeed" if result else "failed"
                print (msg)

    def do_set_angle(self, arg):
        """
        set_angle
        format: write_angle servo_number angle
        servo_number:
            0 bottom servo,
            1 left servo,
            2 right servo,
            3 hand servo
        eg.
        >>> set_angle 0 90
        succeed
        """
        if self.__is_connected():
            values = arg.split(' ')
            if len(values) == 2:
                servo_num = int(values[0])
                angle = float(values[1])
                result = self.uarm.set_servo_angle(servo_num, angle)
                msg = "succeed" if result else "failed"
                print (msg)

    def do_get_angle(self, arg):
        """
        get_angle
        Read current servo angle.
        format: read_angle servo_number
        servo_number:
            0 bottom servo,
            1 left servo,
            2 right servo,
        if no servo_number provide, will list all servos angle
        eg.
        >>> get_angle
        Current Servo Angles: b:17.97, l:112.72, r:17.97, h:151.14
        """
        if self.__is_connected():
            values = arg.split(' ')
            if len(values) == 1:
                if values[0] == '':
                    angles = self.uarm.get_servo_angle()
                    print ("Current Servo Angles: t:{}, l:{}, r:{}, h:{}".format(*angles))

                else:
                    servo_num = int(values[0])
                    angle = self.uarm.get_servo_angle(servo_num)
                    print (angle)

    def do_alert(self, arg):
        """
        alert
        control buzzer
        format: alert frequency duration
        eg.
        alert 1 1
        """
        if self.__is_connected():
            values = arg.split(' ')
            if len(values) == 2:
                frequency = int(values[0])
                duration = float(values[1])
                result = self.uarm.set_buzzer(frequency, duration)
                msg = "succeed" if result else "failed"
                print (msg)

    # def do_set_polar(self, arg):
    #     if self.__is_connected():
    #         values = arg.split(' ')
    #         if len(values) == 4:
    #             result = self.uarm.set_polar_coordinate(values[0], values[1], values[2], values[3])
    #             msg = "succeed" if result else "failed"
    #             print (msg)
    #         if len(values) == 3:
    #             result = self.uarm.set_polar_coordinate(values[0], values[1], values[2])
    #             msg = "succeed" if result else "failed"
    #             print (msg)
    #
    # def do_get_polar(self, arg):
    #     if self.__is_connected():
    #         result = self.uarm.get_polar_coordinate()
    #         if result:
    #             print ("polar coordinate: {}".format(result))

    def do_servo(self, arg):
        """
        Servo status
        format: servo attach servo_number
                servo detach servo_number
        servo_number:
            0 bottom servo,
            1 left servo,
            2 right servo,
            3 hand servo
            all
        eg. servo attach all
            servo detach all
        """
        if self.__is_connected():
            values = arg.split(' ')
            if len(values) == 2:
                if values[0] == 'attach':
                    if values[1] == 'all':
                        self.uarm.set_servo_attach()
                    elif values[1].isdigit():
                        v = int(values[1])
                        if 0 <= v <= 3:
                            self.uarm.set_servo_attach(v)
                elif values[0] == 'detach':
                    if values[1] == 'all':
                        self.uarm.set_servo_detach()
                    elif values[1].isdigit():
                        v = int(values[1])
                        if 0 <= v <= 3:
                            self.uarm.set_servo_detach(v)

    def complete_servo(self, text, line, begidx, endidx):
        if not text:
            completions = self.SERVO_STATUS[:]
        else:
            completions = [f
                           for f in self.SERVO_STATUS
                           if f.startswith(text)
                           ]
        return completions

    def do_serial(self, arg):
        """
        Raw Serial Mode
        You could direct input the communication protocol here.
        """
        if self.__is_connected():
            serial_mode = SerialMode(self.uarm)
            serial_mode.cmdloop()

    # def do_firmware(self, arg):
    #     """
    #     Firmware command
    #     sub command: version, upgrade, force
    #     version - display Remote Firmware version and local Firmware version (if connected)
    #     upgrade - if remote version latter than local firmware version, will flash the latest firmware to uarm
    #     force - force upgrade
    #
    #     """
    #     if arg == 'version':
    #         print ("Remote Firmware Version: {}".format(firmware_helper.get_latest_version()))
    #         if self.__is_connected():
    #             print ("Local Firmware Version: {}".format(self.uarm.firmware_version))
    #     elif arg == 'upgrade':
    #         if self.uarm is not None:
    #             if self.uarm.is_connected():
    #                 self.uarm.disconnect()
    #             subprocess.call(['uarm-firmware','-u', '-p', self.uarm.port])
    #         else:
    #             subprocess.call(['uarm-firmware','-u'])
    #     elif arg == 'force':
    #         if self.uarm is not None:
    #             if self.uarm.is_connected():
    #                 self.uarm.disconnect()
    #             subprocess.call(['uarm-firmware','-d', '-f', '-p', self.uarm.port])
    #         else:
    #             subprocess.call(['uarm-firmware','-df'])
    #     elif arg == '':
    #         print ("please input argument: {}".format(','.join(self.FIRMWARE)))
    #     else:
    #         print ("command not found: {}".format(arg))
    #
    # def complete_firmware(self, text, line, begidx, endidx):
    #     if not text:
    #         completions = self.FIRMWARE[:]
    #     else:
    #         completions = [f
    #                        for f in self.FIRMWARE
    #                        if f.startswith(text)
    #                        ]
    #     return completions

    def do_help(self, arg):
        values = arg.split(' ')
        if len(values) == 1 and values[0] == '':
            help_title = "uArm Command line Help Center"
            help_title += "\n\n"
            help_title += "Please use connect before any control action"
            help_title += "\n"
            help_title += self.help_msg
            print (help_title)
        # print (self.help_msg)
        Cmd.do_help(self,arg)

    def do_quit(self, args):
        """
        Quit, if uarm is connected, will disconnect before quit
        """
        if self.uarm is not None:
            if self.uarm.is_connected():
                self.uarm.disconnect()
        print ("Quiting")
        raise SystemExit

    do_EOF = do_quit


class SerialMode(Cmd):
    prompt = "Serial >>> "
    intro = "Welcome to Serial Mode."
    uarm = None

    def __init__(self, uarm):
        Cmd.__init__(self)
        self.uarm = uarm

    def default(self, line):
        if self.uarm.is_connected():
            response = self.uarm.send_cmd(line)
            print (response)

    def do_quit(self, args):
        return True

    do_EOF = do_quit

def main(args):
    """
    ::

        python -m pyuarm.tools.miniterm
        pyuarm - INFO - pyuarm version: 2.1.1
        pyuarm - INFO - Connecting from port - /dev/cu.usbserial-A600CRE6...
        pyuarm - INFO - connected...
        pyuarm - INFO - Firmware Version: 2.1.4
        Welcome to use uArm Command Line - v0.1.4
        Shortcut:
        Quit: Ctrl + D, or input: quit
        Clear Screen: Ctrl + L

        Input help for more usage
        >>>

    """
    try:
        uarm_cmd = UArmCmd(port=args.port, debug=args.debug)
        uarm_cmd.cmdloop()
    except KeyboardInterrupt:
        print ("KeyboardInterrupt")

if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--port", help="specify port number")
    parser.add_argument("-d", "--debug", help="Open Debug Message")
    args = parser.parse_args()
    main(args)


import argparse
from ..version import __version__
from . import miniterm, list_uarms
from .firmware import flash_firmware
from .calibration import calibrate


def main():
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(dest='cmd')

    parser.add_argument('-v', '--version', action='version', version='pyuarm version: {}'.format(__version__))
    pm = subparsers.add_parser("miniterm")
    pm.add_argument("-p", "--port", help="specify port number")
    pm.add_argument("-d", "--debug", help="Open Debug log", action="store_true")

    pc = subparsers.add_parser("calibrate")
    pc.add_argument("-p", "--port", help="specify port number")
    pc.add_argument("-d", "--debug", help="Open Debug log", action="store_true")
    pc.add_argument("-c", "--check", help="Check the calibrate offset values", action="store_true")

    pl = subparsers.add_parser("list")

    pf = subparsers.add_parser("firmware")
    pf.add_argument("-p", "--port", help="specify port number")
    pf.add_argument("--path", help="firmware path")
    pf.add_argument("-d", "--download", help="download firmware from {}".format(flash_firmware.default_config['download_url']),
                        action="store_true")

    args = parser.parse_args()

    if args.cmd:
        if args.cmd == 'miniterm':
            miniterm.main(args)
        elif args.cmd == 'calibrate':
            calibrate.main(args)
        elif args.cmd == 'list':
            list_uarms.main()
        elif args.cmd == 'firmware':
            flash_firmware.main(args)

if __name__ == '__main__':
    main()


from serial.tools import list_ports

UARM_HWID_KEYWORD = "USB VID:PID=0403:6001"


def uarm_ports():
    uarm_ports = []
    for i in list_ports.comports():
        if i.hwid[0:len(UARM_HWID_KEYWORD)] == UARM_HWID_KEYWORD:
            uarm_ports.append(i[0])
    return uarm_ports


def check_port_plug_in(serial_id):
    ports = list_ports.comports()
    for p in ports:
       if p.serial_number == serial_id:
           return True
    return False


def get_uarm_port_cli():
    uarm_list = uarm_ports()
    ports = uarm_ports()
    if len(ports) > 1:
        i = 1
        for port in ports:
            print ("[{}] - {}".format(i, port))
            i += 1
        port_index = input("Please Choose the uArm Port: ")
        uarm_port = ports[int(port_index) - 1]
        return uarm_port
    elif len(ports) == 1:
        return uarm_list[0]
    elif len(ports) == 0:
        print ("No uArm ports is found.")
        return None


def get_port_property(port_name):
    for p in list_ports.comports():
        if p.device == port_name:
            return p
    return None


def main():
    """
    ::

        $ python -m pyuarm.tools.list_uarms
        /dev/cu.usbserial-A600CVS9
        1 ports found

    """
    ports = uarm_ports()
    for p in ports:
        print (p)
    print ("{0} ports found".format(len(ports)))

if __name__ == '__main__':
    main()



from distutils.util import convert_path
from setuptools import setup, find_packages
import platform

if platform.system() == "Windows":
    package_data = {'pyuarm.tools.calibration': ['calibration.hex'], 'pyuarm.tools.firmware': ['avrdude/windows/*']}
elif platform.system() == "Darwin":
    package_data = {'pyuarm.tools.calibration': ['calibration.hex'], 'pyuarm.tools.firmware': ['avrdude/mac/*']}
else:
    package_data = {'pyuarm.tools.calibration': ['calibration.hex']}

# Get Version from version.py file
main_ns = {}
ver_path = convert_path('pyuarm/version.py')
with open(ver_path) as ver_file:
        exec(ver_file.read(), main_ns)
version = main_ns['__version__']

# Get README.rst for description
long_description = open('README.rst').read()

# Read requirements.txt as install_requires
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(name='pyuarm',
    version=version,
    author='Alex Tan',
    packages=find_packages(),
    entry_points={
            'console_scripts': [
                'uarm_helper = pyuarm.tools.console_scripts:main',
            ]
    },
    # data_files=data_files,
    package_data=package_data,
    include_package_data=True,
    author_email='developer@ufactory.cc',
    url="https://github.com/uarm-developer/pyuarm",
    keywords="pyuarm uarm4py uarmForPython uarm ufactory",
    install_requires=requirements,
    long_description=long_description,
    description='A python library for uArm',
    license='MIT'
)



from __future__ import print_function
import serial
from . import version, protocol, util
from .util import printf, ERROR, DEBUG, UArmConnectException, get_default_logger
from .tools.list_uarms import uarm_ports, get_port_property, check_port_plug_in
from . import PY3
import time


def get_uarm(logger=None):
    """
    ===============================
    Get First uArm Port instance
    ===============================
    It will return the first uArm Port detected by **pyuarm.tools.list_uarms**,
    If no available uArm ports, will print *There is no uArm port available*

    .. raw:python
    >>> import pyuarm
    >>> uarm = pyuarm.get_uarm()
    There is no uArm port available


    :returns: uArm() Instance

    """
    ports = uarm_ports()
    if len(ports) > 0:
        return UArm(port_name=ports[0],logger=logger)
    else:
        printf("There is no uArm port available",ERROR)
        return None

class UArm(object):

    firmware_version = None
    hardware_version = None
    __isConnected = False

    def __init__(self, port_name=None, logger=None, debug=False):
        """
        :param port_name: UArm Serial Port name, if no port provide, will try first port we detect
        :param logger: if no logger provide, will create a logger by default
        :raise UArmConnectException

        UArm port is immediately opened on object creation, if no port provide, we will detect all connected uArm serial
        devices. please reference `pyuarm.tools.list_uarms`
        port is a device name: depending on operating system. eg. `/dev/ttyUSB0` on GNU/Linux or `COM3` on Windows.
        logger will display all info/debug/error/warning messages.
        """
        self.serial_id = 0
        if logger is None:
            util.init_logger(util.get_default_logger(debug))
        else:
            util.init_logger(logger)
        if port_name is None:
            ports = uarm_ports()
            if len(ports) > 0:
                port_name = ports[0]
            else:
                raise UArmConnectException(0, "No uArm ports is found.")
        self.port = get_port_property(port_name)
        self.__serial = serial.Serial(baudrate=115200, timeout=.1)
        self.connect()

    def disconnect(self):
        """
        disconnect will release/close the uarm port
        :return:
        """
        printf("Disconnect from port - {0}...".format(self.port.device))
        self.__serial.close()
        self.__isConnected = False
        self.checking_port_flag = False

    def connect(self):
        """
        This function will open the port immediately. Function will wait for the READY Message for 5 secs. Once received
        READY message, Function will send the Version search command.
        :return:
        """
        try:
            self.__serial.port = self.port.device
            printf("Connecting from port - {0}...".format(self.port.device))
            self.__serial.open()
            timeout_start = time.time()
            timeout = 5
            while time.time() < timeout_start + timeout:
                if self.is_ready():
                    break
            if not self.__isConnected:
                raise UArmConnectException(1, "{} message received timeout.".format(protocol.READY))
        except serial.SerialException as e:
            raise UArmConnectException(0, "port: {}, Error: {}".format(self.port.device, e.strerror))
        self.responseLog = []
        self.get_firmware_version()
        self.get_hardware_version()
        if version.is_a_version(self.firmware_version):
            printf("Firmware Version: {0}".format(self.firmware_version))
            if not version.is_supported_version(self.firmware_version):
                raise UArmConnectException(2,"Firmware Version: {}".format(self.firmware_version))
        else:
            raise UArmConnectException(1, "Firmware Version: {}".format(self.firmware_version))

    def is_connected(self):
        """
        is_connected will return the uarm connected status
        :return: connected status
        """
        try:
            if PY3:
                self.__gen_serial_id()
                cmnd = "#{} {}".format(self.serial_id, protocol.GET_FIRMWARE_VERSION)
                cmndString = bytes(cmnd + "\n", encoding='ascii')
                self.__serial.write(cmndString)
                response = str(self.__serial.readline(),encoding='ascii')
            else:
                self.__gen_serial_id()
                cmnd = "#{} {}".format(self.serial_id, protocol.GET_FIRMWARE_VERSION)
                cmndString = bytes(cmnd + "\n")
                self.__serial.write(cmndString)
                response = self.__serial.readline()
        except serial.serialutil.SerialException:
            self.__isConnected = False
        if self.__serial.isOpen() and self.__isConnected:
            return True
        else:
            return False

    def is_ready(self):
        if PY3:
            ready_msg = bytes(protocol.READY, encoding='ascii')
        else:
            ready_msg = protocol.READY
        if self.__serial.readline().startswith(ready_msg):
            printf("Connected...")
            self.__isConnected = True
            return True
        else:
            return False

    def __gen_serial_id(self):
        if self.serial_id == 999:
            self.serial_id = 0
        else:
            self.serial_id += 1

    def __gen_response_value(self, response):
        if response.startswith(protocol.OK.lower()):
            return response.rstrip().split(' ')[1:]
        else:
            return False


    def __send_and_receive(self, cmnd, timeout=None):
        """
        This command will send a command and receive the uArm response. There must always be a response!
        Responses should be recieved immediately after sending the command, after which the robot will proceed to
        perform the action.
        :param cmnd: a String command, to send to the robot
        :return: The robots response
        """

        if not self.is_connected():
            printf("Communication| Tried to send a command while robot was not connected!")
            return ""

        # Prepare and send the command to the robot
        self.__gen_serial_id()
        cmnd = "#{} {}".format(self.serial_id,cmnd)
        # printf(cmnd, type=ERROR)
        if PY3:
            cmndString = bytes(cmnd + "\n", encoding='ascii')
        else:
            cmndString = bytes(cmnd + "\n")

        try:
            self.__serial.write(cmndString)

        except serial.serialutil.SerialException as e:
            printf("while sending command {}. Disconnecting Serial! \nError: {}".format(cmndString, str(e)),type=ERROR)
            self.__isConnected = False
            return ""

        try:
            if PY3:
                response = str(self.__serial.readline(),encoding='ascii')
            else:
                response = self.__serial.readline()
            if response.startswith("${}".format(self.serial_id)):
                if "E20" in response or "E21" in response:
                    printf("Communication| ERROR: send {}, received error from robot: {}".format(cmndString, response), type=ERROR)
                    return ""
                response = response.replace('\n', '')
                response = response.replace('${} '.format(self.serial_id),'')
                printf("Communication| [{}] {}{}".format(cmnd, " " * (30 - len(cmnd)), response), type=DEBUG)
            else:
                printf("Communication| ERROR: received error from robot: {}".format(response),type=ERROR)
                return ""
            return response.lower()
        except serial.serialutil.SerialException as e:
            printf("while sending command {}. Disconnecting Serial! \nError: {}".format(cmnd,str(e)), type=ERROR)
            self.__isConnected = False
            return ""

    def __parse_cmd(self, message, arguments):
        response_dict = {n: 0 for n in arguments}  # Fill the dictionary with zero's

        # Do error checking, in case communication didn't work
        if message is False:
            printf("UArm.__parse_cmd(): Since an error occurred in communication, returning 0's for all arguments!")
            return response_dict

        # if command not in message:
        #     printf("UArm.__parse_cmd(): ERROR: The message did not come with the appropriate command!")
        #     return response_dict
        #
        # # Get rid of the "command" part of the message, so it's just arguments and their numbers
        # message = message.replace(command, "")

        # Get the arguments and place them into the array
        for i, arg in enumerate(arguments):
            if i < len(arguments) - 1:
                response_dict[arg] = message[message.find(arg) + 1: message.find(arguments[i + 1])]
            else:
                response_dict[arg] = message[message.find(arg) + 1:]

            response_dict[arg] = float(response_dict[arg])

        return response_dict

# -------------------------------------------------------- Commands ----------------------------------------------------

    def get_firmware_version(self):
        """
        Get the firmware version.
        Protocol Cmd: `protocol.GET_FIRMWARE_VERSION`
        :return: firmware version, if failed return False
        """
        cmd = protocol.GET_FIRMWARE_VERSION
        response = self.__send_and_receive(cmd)

        value = self.__gen_response_value(response)
        if value:
            self.firmware_version = value[0][1:]
        else:
            return False

    def get_hardware_version(self):
        """
        Get the Product version.
        Protocol Cmd: `protocol.GET_HARDWARE_VERSION`
        :return: firmware version, if failed return False
        """
        cmd = protocol.GET_HARDWARE_VERSION
        response = self.__send_and_receive(cmd)

        value = self.__gen_response_value(response)
        if value:
            self.hardware_version = value[0][1:]
        else:
            return False

    def get_simulation(self, x, y, z):
        """
        validate the coordinate (x,y,z) if it can be reached or not.
        :param x:
        :param y:
        :param z:
        :return:
        """
        x = str(round(x, 2))
        y = str(round(y, 2))
        z = str(round(z, 2))
        cmd = protocol.SIMULATION.format(x, y, z)
        response = self.__send_and_receive(cmd)
        value = self.__gen_response_value(response)
        if value:
            return value
        else:
            return False

    def set_position(self, x, y, z, speed=300, relative=False):
        """
        Move uArm to the position (x,y,z) unit is mm, speed unit is mm/sec
        :param x:
        :param y:
        :param z:
        :param speed:
        :return:
        """
        x = str(round(x, 2))
        y = str(round(y, 2))
        z = str(round(z, 2))
        s = str(round(speed, 2))
        if relative:
            command = protocol.SET_POSITION_RELATIVE.format(x, y, z, s)
        else:
            command = protocol.SET_POSITION.format(x, y, z, s)
        response = self.__send_and_receive(command)
        if response.startswith(protocol.OK.lower()):
            return True
        else:
            return False


    def set_servo_angle(self, servo_number, angle):
        """
        Set uArm Servo Angle, 0 - 180 degrees, this Function will include the manual servo offset.
        :param servo_number: lease reference protocol.py SERVO_BOTTOM, SERVO_LEFT, SERVO_RIGHT, SERVO_HAND
        :param angle: 0 - 180 degrees
        :return: succeed True or Failed False
        """
        cmd = protocol.SET_ANGLE.format(str(servo_number), str(angle))
        response = self.__send_and_receive(cmd)
        if response.startswith(protocol.OK.lower()):
            return True
        else:
            return False


    # def set_servo_raw_angle(self, servo_number, angle):
    #     """
    #     Set uArm Servo Raw Angle, 0 - 180 degrees, this Function will exclude the manual servo offset.
    #     :param servo_number: lease reference protocol.py SERVO_BOTTOM, SERVO_LEFT, SERVO_RIGHT, SERVO_HAND
    #     :param angle: 0 - 180 degrees
    #     :return: succeed True or Failed False
    #     """
    #     cmd = protocol.SET_RAW_ANGLE.format(str(servo_number), str(angle))
    #     response = self.__send_and_receive(cmd)
    #     if response.startswith("s"):
    #         return True
    #     else:
    #         return False

    def set_wrist(self, angle):
        """
        Set uArm Hand Wrist Angle. Include servo offset.
        :param angle:
        :return:
        """
        return self.set_servo_angle(protocol.SERVO_HAND, angle)

    def set_pump(self, ON):
        """
        Control uArm Pump On or OFF
        :param ON: True On, False OFF
        :return: succeed True or Failed False
        """
        cmd = protocol.SET_PUMP.format(1 if ON else 0)
        response = self.__send_and_receive(cmd)
        if response.startswith(protocol.OK.lower()):
            return True
        else:
            return False

    def set_servo_attach(self, servo_number=None):
        """
        Set Servo status attach, Servo Attach will lock the servo, You can't move uArm with your hands.
        :param servo_number: If None, will attach all servos, please reference protocol.py SERVO_BOTTOM, SERVO_LEFT, SERVO_RIGHT, SERVO_HAND
        :return: succeed True or Failed False
        """
        if servo_number is not None:
            cmd = protocol.ATTACH_SERVO.format(servo_number)
            response = self.__send_and_receive(cmd)
            if response.startswith(protocol.OK.lower()):
                return True
            else:
                return False
        else:
            if self.set_servo_attach(0) and self.set_servo_attach(1) \
                    and self.set_servo_attach(2) and self.set_servo_attach(3):
                return True
            else:
                return False

    def set_servo_detach(self, servo_number=None):
        """
        Set Servo status detach, Servo Detach will unlock the servo, You can move uArm with your hands. But move function won't be effect until you attach.
        :param servo_number: If None, will detach all servos, please reference protocol.py SERVO_BOTTOM, SERVO_LEFT, SERVO_RIGHT, SERVO_HAND
        :return: succeed True or Failed False
        """
        if servo_number is not None:
            cmd = protocol.DETACH_SERVO.format(servo_number)
            response = self.__send_and_receive(cmd)
            if response.startswith(protocol.OK.lower()):
                return True
            else:
                return False
        else:
            if self.set_servo_detach(0) and self.set_servo_detach(1) \
                    and self.set_servo_detach(2) and self.set_servo_detach(3):
                return True
            else:
                return False

    def set_buzzer(self, frequency, duration):
        """
        Turn on the uArm Buzzer
        :param frequency: The frequency, in Hz
        :param duration: The duration of the buzz, in seconds
        :return:
        """
        cmd = protocol.SET_BUZZER.format(frequency, duration)
        response = self.__send_and_receive(cmd)
        if response.startswith(protocol.OK.lower()):
            return True
        else:
            return False

    def get_position(self):
        """
        Get Current uArm position (x,y,z) mm
        :return: Returns an array of the format [x, y, z] of the robots current location
        """
        response = self.__send_and_receive(protocol.GET_COOR)
        value = self.__gen_response_value(response)
        if value:
            parse_cmd = self.__parse_cmd(response, ["x", "y", "z"])
            coordinate = [parse_cmd["x"], parse_cmd["y"], parse_cmd["z"]]
            return coordinate
        else:
            return False

    def is_moving(self):
        """
        Detect is uArm moving
        :return: Returns a 0 or a 1, depending on whether or not the robot is moving.
        """
        response = self.__send_and_receive(protocol.GET_IS_MOVE)
        value = self.__gen_response_value(response)
        if value:
            if value[1:] == "1":
                return True
            else:
                return False
        else:
            return False

    def set_gripper(self, catch):
        """
        Turn On/Off Gripper
        :param catch: True On/ False Off
        :return:
        """
        cmd = protocol.SET_GRIPPER.format(1 if catch else 0)
        response = self.__send_and_receive(cmd)
        if response.startswith(protocol.OK.lower()):
            return True
        else:
            return False

    # def set_polar_coordinate(self, rotation, stretch, height, speed=100):
    #     """
    #     Polar Coordinate, rotation, stretch, height.
    #     :param rotation:
    #     :param stretch:
    #     :param height:
    #     :param speed:
    #     :return:
    #     """
    #     cmd = protocol.SET_POLAR.format(stretch, rotation, height, speed)
    #     if self.__send_and_receive(cmd).startswith("s"):
    #         return True
    #     else:
    #         return False
    #
    # def get_polar_coordinate(self):
    #     cmd = protocol.GET_POLAR
    #     response = self.__send_and_receive(cmd)
    #     if response.startswith("s"):
    #         parse_cmd = self.__parse_cmd(response[1:], ["s", "r", "h"])
    #         polar_crd = [parse_cmd["s"], parse_cmd["r"], parse_cmd["h"]]
    #         return polar_crd
    #     else:
    #         return False

    def get_servo_angle(self, servo_number=None):
        """
        Get Current uArm Servo Angles include offset
        :param servo_number: if None, Return 4 servos Current Angles
        :return: Returns an angle in degrees, of the servo
        """
        cmd = protocol.GET_ANGLE

        response = self.__send_and_receive(cmd)

        value = self.__gen_response_value(response)
        if value:
            parse_cmd = self.__parse_cmd(response, ["b", "l", "r", "h"])
            angles = [parse_cmd["b"], parse_cmd["l"], parse_cmd["r"], parse_cmd["h"]]
            if servo_number is not None:
                if 0 <= servo_number <= 3:
                    return angles[servo_number]
                else:
                    return False
            else:
                return angles
        else:
            return False

    # def get_servo_raw_angle(self, servo_number=None):
    #     """
    #     Get Current uArm Servo Angles exclude offset
    #     :param servo_number:  if None, Return 4 servos Current Angles
    #     :return: Returns an angle in degrees, of the servo
    #     """
    #     cmd = protocol.GET_RAW_ANGLE
    #
    #     response = self.__send_and_receive(cmd)
    #     if response.startswith("s"):
    #         parse_cmd = self.__parse_cmd(response[1:],  ["b", "l", "r", "h"])
    #         angles = [parse_cmd["b"], parse_cmd["l"], parse_cmd["r"], parse_cmd["h"]]
    #         if servo_number is not None:
    #             if 0 <= servo_number <= 3:
    #                 return angles[servo_number]
    #             else:
    #                 return False
    #         else:
    #             return angles
    #     else:
    #         return False

    def get_tip_sensor(self):
        """
        Get Status from Tip Sensor
        :return: True On/ False Off
        """
        response = self.__send_and_receive(protocol.GET_TIP_SENSOR)
        value = self.__gen_response_value(response)

        if value:
            if "".join(value)[1:] == "0":
                return True
            else:
                return False
        else:
            return False

    def get_rom_data(self, address, data_type=protocol.EEPROM_DATA_TYPE_BYTE):
        """
        Get DATA From EEPROM
        :param address: 0 - 2048
        :param data_type: EEPROM_DATA_TYPE_FLOAT, EEPROM_DATA_TYPE_INTEGER, EEPROM_DATA_TYPE_BYTE
        :return:
        """
        cmd = protocol.GET_EEPROM.format(address, data_type)
        response = self.__send_and_receive(cmd)
        value = self.__gen_response_value(response)
        if value:
            # print("val: {}".format(type)))
            val = "".join(value)[1:]

            if data_type == protocol.EEPROM_DATA_TYPE_FLOAT:
                return float(val)
            elif data_type == protocol.EEPROM_DATA_TYPE_INTEGER:
                return int(val)
            elif data_type == protocol.EEPROM_DATA_TYPE_BYTE:
                return int(val)
        else:
            return False


    def set_rom_data(self, address, data, data_type=protocol.EEPROM_DATA_TYPE_BYTE):
        """
        Set DATA to EEPROM
        :param address: 0 - 2048
        :param data: Value
        :param data_type: EEPROM_DATA_TYPE_FLOAT, EEPROM_DATA_TYPE_INTEGER, EEPROM_DATA_TYPE_BYTE
        :return:
        """
        cmd = protocol.SET_EEPROM.format(address, data_type, data)
        response = self.__send_and_receive(cmd)
        if response.startswith(protocol.OK.lower()):
            return True
        else:
            return False

    def get_sn_number(self):
        """
        Get Serial Number
        :return:
        """
        serial_number = ""
        for i in range(15):
            serial_number += chr(int(self.get_rom_data(protocol.SERIAL_NUMBER_ADDRESS + i)))
        return serial_number

    def get_analog(self,pin):
        cmd = protocol.GET_ANALOG.format(pin)
        response = self.__send_and_receive(cmd)
        if response.startswith("s"):
            val = response[1:]
            return int(float(val))

    def get_digital(self,pin):
        cmd = protocol.GET_DIGITAL.format(pin)
        response = self.__send_and_receive(cmd)
        if response.startswith("s"):
            val = response[1:]
            return val
